local addonName, addon = ...

-----------------------------------------------------------------------
-- Locale
--

local L = addon.L

local HELP = { L.HELP1, L.HELP2, L.HELP3, L.HELP4 }
local MENU = {
	NOTE = L.MENU_NOTE,
	NOTE_DESC = L.MENU_NOTE_DESC,
	IGNORE = L.MENU_IGNORE,
	IGNORE_DESC = L.MENU_IGNORE_DESC,
}

-----------------------------------------------------------------------
-- Tooltip scanner, mostly stolen from Gratuity.
--

local scanItem
do
	local scanner = CreateFrame("GameTooltip")
	scanner:SetOwner(WorldFrame, "ANCHOR_NONE")
	local lcache, rcache = {}, {}
	for i = 1, 30 do
		lcache[i], rcache[i] = scanner:CreateFontString(), scanner:CreateFontString()
		lcache[i]:SetFontObject(GameFontNormal); rcache[i]:SetFontObject(GameFontNormal)
		scanner:AddFontStrings(lcache[i], rcache[i])
	end
	local function erase()
		for i = 1, scanner:NumLines() do rcache[i]:SetText() end
		scanner:ClearLines()
	end
	local scannerCache = {}
	function scanItem(itemLink)
		erase()
		scanner:SetHyperlink(itemLink)
		wipe(scannerCache)
		for i = 1, scanner:NumLines() do
			local left, right = lcache[i]:GetText(), rcache[i]:GetText()
			if left then scannerCache[#scannerCache + 1] = left end
			if right then scannerCache[#scannerCache + 1] = right end
		end
		return table.concat(scannerCache)
	end
end

-----------------------------------------------------------------------
-- CraftyGnome
--

local bangMappings = {
	alchemy = GetSpellInfo(2259),
	blacksmith = GetSpellInfo(2018),
	cooking = GetSpellInfo(2550),
	leather = GetSpellInfo(2108),
	enchanting = GetSpellInfo(7411),
	engineering = GetSpellInfo(4036),
	firstaid = GetSpellInfo(746),
	jewel = GetSpellInfo(25229),
	tailor = GetSpellInfo(3908),
	inscription = GetSpellInfo(45357),
}
bangMappings.alc = bangMappings.alchemy
bangMappings.bs = bangMappings.blacksmith
bangMappings.cook = bangMappings.cooking
bangMappings.lw = bangMappings.leather
bangMappings.enchant = bangMappings.enchanting
bangMappings.eng = bangMappings.engineering
bangMappings.aid = bangMappings.firstaid
bangMappings.jc = bangMappings.jewel
bangMappings.gem = bangMappings.jewel
bangMappings.tailoring = bangMappings.tailor
bangMappings.ins = bangMappings.inscription
bangMappings.craft = true
bangMappings.cg = true
bangMappings.next = true
bangMappings.help = true

local MAX_RESULTS = 5

local db = nil
local ignoreDB = nil
local notesDB = nil

local bop = ITEM_BIND_ON_PICKUP
local cache = {}
local acceptedCommands = {}
local player = UnitName("player")
local eventFrame = CreateFrame("Frame")

--------------------------------------------------------------------------------
-- Right-click menu for craft/tradeskill UI
--

local function forceUpdate()
	if TradeSkillFrame and TradeSkillFrame:IsVisible() then TradeSkillFrame_Update() end
end

local currentItem = nil

local function isIgnored(item)
	if type(item) == "table" then
		return ignoreDB[currentItem]
	else
		return item == "tradeSkillListLink" or ignoreDB[item]
	end
end
local function toggleIgnored()
	ignoreDB[currentItem] = not ignoreDB[currentItem]
	forceUpdate()
end

local function getNote(item)
	return notesDB[item or currentItem]
end
local function setNote(item, note)
	if type(note) ~= "string" or note:trim():len() < 2 then note = nil end
	notesDB[item or currentItem] = note
	forceUpdate()
end
local function showNotePopup()
	StaticPopupDialogs["CraftyGnomeNote"].text = L.POPUP_TEXT:format(currentItem)
	StaticPopup_Show("CraftyGnomeNote")
end

local menu = {{
	text = MENU.NOTE,
	tooltipTitle = MENU.NOTE,
	tooltipText = MENU.NOTE_DESC,
	tooltipOnButton = 1,
	notCheckable = 1,
	func = showNotePopup,
},{
	text = MENU.IGNORE,
	tooltipTitle = MENU.IGNORE,
	tooltipText = MENU.IGNORE_DESC,
	tooltipOnButton = 1,
	func = toggleIgnored,
	checked = isIgnored,
},}

local popupFrame = CreateFrame("Frame", "CraftyGnomeMenu", UIParent, "UIDropDownMenuTemplate")
local function openMenu(frame, recipe)
	StaticPopup_Hide("CraftyGnomeNote")
	currentItem = recipe
	EasyMenu(menu, popupFrame, frame, 20, 4, "MENU")
end

local function onTradeSkillDoubleClick(self)
	local recipe = GetTradeSkillRecipeLink(self:GetID())
	if recipe then openMenu(self, recipe) end
end
local recursive = nil
local function updateSkillLine(line, text, recipe)
	local note = getNote(recipe)
	local ignored = isIgnored(recipe)
	local append = nil
	if note then append = " - " .. note end
	if ignored then append = (append or "") .. L.IGNORED end
	if append then
		recursive = true
		line:SetText(text .. append)
	end
end
local function onTradeSkillLineSet(self, text)
	if recursive then recursive = nil; return end
	local recipe = GetTradeSkillRecipeLink(self:GetID())
	if recipe then updateSkillLine(self, text, recipe) end
end
function eventFrame:ADDON_LOADED(name)
	if name == "Blizzard_TradeSkillUI" then
		for i = 1, 8 do
			local button = _G["TradeSkillSkill" .. i]
			button:SetScript("OnDoubleClick", onTradeSkillDoubleClick)
			hooksecurefunc(button, "SetText", onTradeSkillLineSet)
		end
		print(L.MENU_HELP)
	end
end

--------------------------------------------------------------------------------
-- Craft/tradeskill scanning
--

-- XXX we should just save the spell ID in the database like I did originally,
-- XXX maybe we can use GetSpellLink and GetSpellInfo on the recipe IDs in 2.4
-- XXX in any case, we can just create the links manually from the ID like
-- XXX |cffffd000|Henchant:"..recipeId.."|h[Placeholder Name]|h|r"
-- XXX and then use tooltip:SetHyperlink
-- XXX if we do this we also have to get the name and from the tooltip so we can
-- XXX construct the recipe link properly when someone wants a link

function eventFrame:TRADE_SKILL_SHOW()
	local isLink, unit = IsTradeSkillLinked()
	if unit and unit ~= player then return end
	local name = GetTradeSkillLine()
	if not name then return end
	if not db[name] then db[name] = {} end
	db[name].tradeSkillListLink = GetTradeSkillListLink()
	for i = 1, GetNumTradeSkills() do
		local skillName, skillType = GetTradeSkillInfo(i)
		if skillType ~= "header" and skillName then
			local recipe = GetTradeSkillRecipeLink(i)
			if recipe and not db[name][recipe] and not isIgnored(recipe) then
				local link = GetTradeSkillItemLink(i)
				local itemType = GetItemFamily(link)
				if link then
					local _, _, _, level = GetItemInfo(link)
					local l = level or 1
					if not db[name][recipe] or db[name][recipe] < l then
						db[name][recipe] = l
					end
				end
			end
		end
	end
end

--------------------------------------------------------------------------------
-- Query processing
--

do
	local ctl = ChatThrottleLib
	local reply = ctl and
		function(msg, player) ctl:SendChatMessage("BULK", "CraftyGnome", L.PREFIX .. msg, "WHISPER", nil, player) end or
		function(msg, player) SendChatMessage(L.PREFIX .. msg, "WHISPER", nil, player) end

	local function updateCache(skill)
		if not cache[skill] then cache[skill] = {} end
		for item in pairs(db[skill]) do
			if not cache[skill][item] and not isIgnored(item) then
				local text = scanItem(item)
				if not text:find(bop) then
					cache[skill][item] = text:lower()
				end
			end
		end
	end

	local function updateAcceptedCommands()
		wipe(acceptedCommands)
		for k, v in pairs(bangMappings) do
			if db[v] then
				acceptedCommands[#acceptedCommands+1] = "!" .. k
			end
		end
		acceptedCommands[#acceptedCommands+1] = "!help"
		acceptedCommands[#acceptedCommands+1] = "!craft"
		acceptedCommands[#acceptedCommands+1] = "!cg"
	end

	local function search(skill, ...)
		local tbl = nil
		if cache[skill] then
			local c = select("#", ...)
			for item, details in pairs(cache[skill]) do
				if not isIgnored(item) then
					local matchAll = true
					for i = 1, c do
						local token = tostring(select(i, ...)):lower()
						if not details:find(token) then
							matchAll = nil
						end
					end
					if matchAll then
						if not tbl then tbl = {} end
						tbl[#tbl + 1] = item
					end
				end
			end
		end
		return tbl
	end

	local function replyWithNote(item, player)
		local note = getNote(item)
		if note then
			reply(L.NOTE:format(item, note), player)
		else
			reply(item, player)
		end
	end
	local moreResults = {}
	local function replyResults(tbl, player)
		if not tbl then return end
		if #tbl > MAX_RESULTS then
			for i = 1, MAX_RESULTS do
				if tbl[i] then replyWithNote(tbl[i], player) end
			end
			for i = 1, MAX_RESULTS do table.remove(tbl, 1) end
			reply(L.TOO_MANY:format(#tbl), player)
		else
			for i, v in next, tbl do replyWithNote(v, player) end
			tbl = nil
		end
		return tbl
	end

	local bangHandlers = {}
	bangHandlers.next = function(player)
		moreResults[player] = replyResults(moreResults[player], player)
	end
	bangHandlers.craft = function(player)
		updateAcceptedCommands()
		reply(ACCEPTED:format(table.concat(acceptedCommands, ", ")), player)
	end
	bangHandlers.cg = bangHandlers.craft
	bangHandlers.help = function(player, isWhisper)
		if not isWhisper then return end
		for i, v in next, HELP do
			reply(v, player)
		end
		bangHandlers.craft(msg, player)
	end

	local skillSorters = {}
	local function processChat(msg, player, isWhisper)
		local cmd = select(3, msg:find("^!(%S+)"))
		if cmd then cmd = cmd:lower() end
		if not cmd or not bangMappings[cmd] then return end
		if bangHandlers[cmd] then
			bangHandlers[cmd](player, isWhisper)
			return
		end
		local query = select(3, msg:find("^!%S+%s(.*)"))
		local skill = bangMappings[cmd]
		if query then
			if db[skill] then
				updateCache(skill)

				local name = GetItemInfo(query)
				local results = nil
				if name then
					-- This is an item, don't split it up into multiple tokens
					results = search(skill, name)
				else
					results = search(skill, strsplit(" ", query))
				end
				if results then
					if not skillSorters[skill] then
						skillSorters[skill] = function(a, b)
							return db[skill][a] > db[skill][b]
						end
					end
					table.sort(results, skillSorters[skill])
					moreResults[player] = replyResults(results, player)
				elseif isWhisper then
					reply(L.NO_DATA:format(query), player)
				end
			else
				updateAcceptedCommands()
				reply(L.NO_SKILL:format(table.concat(acceptedCommands, ", ")), player)
			end
		elseif isWhisper then
			if db[skill] and db[skill].tradeSkillListLink then
				reply(db[skill].tradeSkillListLink, player)
			end
			reply(L.CG_USAGE:format(cmd), player)
		end
	end
	function eventFrame:CHAT_MSG_GUILD(msg, player)
		processChat(msg, player)
	end
	function eventFrame:CHAT_MSG_WHISPER(msg, player)
		processChat(msg, player, true)
	end
end

--------------------------------------------------------------------------------
-- Core addon and event handling
--

function eventFrame:PLAYER_LOGIN()
	if not CraftyGnomeDB then CraftyGnomeDB = {} end
	local realm = GetRealmName()
	if not CraftyGnomeDB[realm] then CraftyGnomeDB[realm] = {} end
	local faction = UnitFactionGroup("player")
	if not CraftyGnomeDB[realm][faction] then CraftyGnomeDB[realm][faction] = {} end
	db = CraftyGnomeDB[realm][faction]

	if not db.version or db.version < 5 then
		wipe(db)
		db.version = 5
	end

	if not CraftyGnomeDB.ignoredItems then CraftyGnomeDB.ignoredItems = {} end
	ignoreDB = CraftyGnomeDB.ignoredItems

	if not CraftyGnomeDB.notes then CraftyGnomeDB.notes = {} end
	notesDB = CraftyGnomeDB.notes

	StaticPopupDialogs["CraftyGnomeNote"] = {
		text = nil,
		button1 = MENU.NOTE,
		button2 = CANCEL,
		whileDead = 1,
		hideOnEscape = 1,
		timeout = 0,
		OnShow = function(self)
			self.editBox:SetText(getNote(currentItem) or "")
		end,
		EditBoxOnEnterPressed = function(self)
			setNote(currentItem, self:GetText())
			self:GetParent():Hide()
		end,
		EditBoxOnEscapePressed = function(self)
			self:GetParent():Hide()
		end,
		OnAccept = function(self)
			setNote(currentItem, self.editBox:GetText())
		end,
		hasEditBox = 1,
	}
end

ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", function(self, event, msg)
	local cmd = select(3, msg:find("^!(%S+)"))
	if cmd and bangMappings[cmd:lower()] then return true end
end)
local chatPrefixTest = "^" .. L.PREFIX
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", function(self, event, msg)
	if msg:find(chatPrefixTest) then return true end
end)

eventFrame:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)
eventFrame:RegisterEvent("PLAYER_LOGIN")
eventFrame:RegisterEvent("CHAT_MSG_GUILD")
eventFrame:RegisterEvent("CHAT_MSG_WHISPER")
eventFrame:RegisterEvent("TRADE_SKILL_SHOW")
eventFrame:RegisterEvent("ADDON_LOADED")

